!
! Module: evigmap_mod
!
! Authors:  Ian Stewart - SSC
!          
! Description: Makes a map in sky coords of the vignetting function.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
!

module evigmap_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use cal
  use dal
  use param  
  use errorHandling
  use dss
  use caloaldefs	! for SpacecraftAttitudeType

  use exposure_map

  use binned_att
  use cube_aux		! for intervalVecType
  use eimchip2sky_aux	! for writeCubeInSkyCoords.

use array_utils
  use cal_aux
  use dal_aux		! for readArrayData
!  use dss_aux		! for readGtisFromDss
  use exposure
!  use save_image
  use wcs_aux
!use dump

  implicit none 

  private 

  public evigmap

  ! Command-line parameters:
  character(256) ::&
    dssSetName = '',&
    templateImageSetName = '',&
    eventSetName = '',&
    expMapCubeSetName='',&
!!!    spectrumSetTabName = '',&
    binnedAttSetName = '',&
    vigMapSetName = '',&
    vigMapCubeSetName='',&
    tempImageSetname=''
  character(20) ::&
    outputStyle = '',&
    attitudeStyle = ''
!!!  logical(bool) ::&
!!!    withSpectrumSet = .false.

!  ! Global variables (these also defined in lccorr_include):
!  integer(int16), parameter ::&
!    numMicropixelDivs = 5 ! used in call to calcCcdExpMap.
!  logical(bool), parameter ::&
!    pleaseCorrectDither = .true. ! used in call to calcCcdExpMap.
  integer(int32), parameter ::&
    numESamples = 50
!, numThetaSamples = 50, numPhiSamples = 36

contains 

  !---------------------------------------------------------------------
  subroutine evigmap
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: expMapCube(:,:,:)
    type(WcsT) :: wcs
    type(DataSetT) :: dssSet
    type(BlockT) :: dssBlock
    type(DssT) :: dssPointer
    integer(int32) :: instrumentId, filtId
    real(double) :: expStartTimeStamp, EeVsamples(numESamples)
    type(SpacecraftAttitudeType) :: scAttitude, instrumAttitude
    type(localTAWtype) :: timeAttWcs
    integer(int16) ::&
      chipMinRawX, chipMaxRawX, chipMinRawY, chipMaxRawY
    integer(int8) ::&
      numCcds, numNodes, ccdNum, nodeNum, ccdNodeNum, ccdnr, maxCcdNodeNum
    integer(int8), allocatable :: ccd2ccdnr(:,:)
    character(2), allocatable :: ccdStr(:,:)
    type(expInfoType), allocatable :: expInfo(:,:)
    type(RangeT) :: EeVrange
    real(single) ::&
      spectrum(numESamples), filterTrans(numESamples), effArea(numESamples),&
      EeVFunctions(numESamples), eNorm
    real(single), allocatable :: ccdVigMaps(:,:,:)
    type(intervalVecType), allocatable :: gtiVec(:)

    !--------------------------------------------------------------------
    ! Start:

    call readParameters

    call readArrayData(expMapCubeSetName, expMapCube) ! in ssclib/dal_aux.

    !....................................................................
    ! DSS and cal setups:

    call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.

    dssSet = dataSet(dssSetName, READ)
    dssBlock = block(dssSet, 0)
    dssPointer = dataSubspace(dssBlock)

    call getCalInfo(dssSet, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! in ssclib/cal_aux.

    timeAttWcs%timeStamp = expStartTimeStamp
    timeAttWcs%attitude  = scAttitude
    timeAttWcs%wcs       = wcs

    if (instrumentId .eq. EMOS1&
    .or. instrumentId .eq. EMOS2) then
      chipMinRawX = EMOS_MIN_X
      chipMaxRawX = EMOS_MAX_X
      chipMinRawY = EMOS_MIN_Y
      chipMaxRawY = EMOS_MAX_Y
      numCcds = 7
      numNodes = 2
!      numCanonicalPatterns = 13
    else ! EPN.
      chipMinRawX = EPN_MIN_X
      chipMaxRawX = EPN_MAX_X
      chipMinRawY = EPN_MIN_Y
      chipMaxRawY = EPN_MAX_Y
      numCcds = 12
      numNodes = 1
!      numCanonicalPatterns = 5
    end if

    allocate(ccd2ccdnr(numCcds, numNodes))
    allocate(ccdStr(numCcds, numNodes))

    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
        ccdnr = 10 * (nodeNum - 1) + ccdNum
        ccd2ccdnr(ccdNum, nodeNum) = ccdnr
        write(ccdStr(ccdNum, nodeNum), '(i2.2)') ccdnr
      end do
    end do

    !....................................................................
    ! Get info from event list:

    allocate(expInfo(numCcds, numNodes))
    call readCcdExposureInfo(eventSetName, ccdStr, instrumentId&
      , expInfo, filtId) ! in exposure_map

    call CAL_setState(filterId=filtId)

    !....................................................................
    ! Get energy/spectral/exposure info (similar bit of code to
    ! lccorr_corrector):

!!!    if (!** with spectrum) then
!!!      call getEnergyBounds(dssPointer, dssBlock, expInfo(1,1)%temperature&
!!!        , expStartTimeStamp, EeVrange, spectrum)
!!!    else
      call getEnergyBounds(dssPointer, dssBlock, expInfo(1,1)%temperature&
        , expStartTimeStamp, EeVrange) ! in exposure_map.
!!!    end if

    call release(dssSet) ! need it until now for the dss.
    call sampleEnergyRange(EeVrange, EeVsamples) ! in ssclib/exposure.

!!!    if (!** with spectrum) then
!*** interpolate user spectrum
!!!    else
      call constructModelSpectrum(EeVsamples, spectrum) ! in exposure_map.
!!!    end if

    call getFilterTransmission(EeVsamples, filterTrans) ! in exposure.f90
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure.
    EeVFunctions = spectrum * effArea * filterTrans

    !....................................................................

    maxCcdNodeNum = 10 * (numNodes - 1) + numCcds
    allocate(ccdVigMaps(maxCcdNodeNum, chipMinRawX:chipMaxRawX&
      , chipMinRawY:chipMaxRawY))
    allocate(gtiVec(maxCcdNodeNum))
    call readGtisFromCube(expMapCubeSetName, gtiVec)
    ! in chipcubelib/cube_aux.

    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
      if (.not. expInfo(ccdNum, nodeNum)%exposuExists) cycle

        if (instrumentId == EPN) then ! just to stop the cal giving a warning:
          call CAL_setState(ccdChipId=int(ccdNum, int32))
        else
          call CAL_setState(ccdChipId=int(ccdNum, int32)&
            , ccdNodeId=int(nodeNum - 1, int32))
        end if

!*** set mode id

        ccdNodeNum = ccd2ccdnr(ccdNum, nodeNum)
        call message(AppMsg, Verbose, "********************************")
        call message(AppMsg, Verbose, "Found EXPOSUnn extension for &
          &CCD/node number nn = "//ccdStr(ccdNum, nodeNum))

        eNorm = trapezoidSum(real(EeVsamples, single)&
          , spectrum * effArea * filterTrans)
        ! in ssclib/array_utils.

!!!        call calcVigMap(EeVsamples, EeVfunctions, (/EeVrange/)&
!!!          , ccdVigMaps(ccdNodeNum,:,:))
ccdVigMaps(ccdNodeNum,:,:) = 1.0

        ccdVigMaps(ccdNodeNum,:,:) = ccdVigMaps(ccdNodeNum,:,:) / eNorm

print *, 'max(vig), max(exp):', maxval(ccdVigMaps(ccdNodeNum,:,:)), maxval(expMapCube(ccdNodeNum,:,:))
        ccdVigMaps(ccdNodeNum,:,:) = ccdVigMaps(ccdNodeNum,:,:)&
          * expMapCube(ccdNodeNum,:,:)
      end do
    end do

!    call release(dssSet) ! need it until now for the dss.

    deallocate(expMapCube)

    !....................................................................

    select case(outputStyle)
    case('sky')
      call writeCubeInSkyCoords(ccdVigMaps&
        , reshape(ccd2ccdnr, (/maxCcdNodeNum/)), expStartTimeStamp, wcs&
        , gtiVec, templateImageSetName, dssSetName, vigMapSetName&
        , tempImageSetName, 'rate', .true., attitudeStyle, binnedAttSetName)
      ! in chipcubelib/eimchip2sky_aux.

    case('raw')
      call writeCubeData(vigMapCubeSetName, templateImageSetName, ccdVigMaps)
      call writeGtisToCube(vigMapCubeSetName, gtiVec)
      ! both in chipcubelib/cube_aux.
    case default
call error('',"")
    end select

    ! Deallocate all the gtis:

    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
        ccdNodeNum = ccd2ccdnr(ccdNum, nodeNum)
        deallocate(gtiVec(ccdNodeNum)%gtis)
      end do
    end do

  end subroutine evigmap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    dssSetName = stringParameter('dssset')
    templateImageSetName = stringParameter('templateset')
    eventSetName = stringParameter('eventset')
    expMapCubeSetName = stringParameter('expcubeset')
!!!    withSpectrumSet = booleanParameter('withspectrumset')
!!!    if (withSpectrumSet)&
!!!      spectrumSetTabName = stringParameter('spectrumset')
    outputStyle = stringParameter('outputstyle')

    select case(outputStyle)
    case('sky')
      vigMapSetName = stringParameter('vigimgset')
    case('raw')
      vigMapCubeSetName = stringParameter('vigcubeset')
    case default
call error('',"")
    end select

    attitudeStyle = stringParameter('attstyle')
    select case(attitudeStyle)
    case('binnedset')
      binnedAttSetName = stringParameter('binnedattset')
    case('template')
call error('notYetSupported',"") !** just for the present.
    case default
call error('',"")
    end select

!    tempImageSetname = stringParameter('tempset')
tempImageSetname = 'temp.ds'

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module evigmap_mod

