!
! Module: etimedither_mod
!
! Authors:  Ian Stewart - SSC
!          
! Description: This task randomizes PN event times within the frame duration.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $

module etimedither_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use param  
  use errorHandling

  use dal_aux
  use random_aux
  use ssc_misc

  implicit none 

  private 

  public etimedither

contains 

  !---------------------------------------------------------------------
  subroutine etimedither
  !---------------------------------------------------------------------

    implicit none

    type frameType
      real(double) :: time, timeDel
    end type frameType
    type allFramesType
      logical :: ccdIsPresent
      type(frameType), pointer :: frames(:)
    end type allFramesType

    character(256) :: eventSetName=''
    type(DataSetT) :: set
    type(TableT) :: eventTab, exposuTab
    integer ::&
      numEvents, minCcdNodeNum, maxCcdNodeNum, trialCcdNodeNum, ccdNodeNum,&
      eventNum, startFrameNum, bestFrameNum, frameNum, approxNumFramesAdvanced
    real(single), allocatable :: rands(:)
    integer(int8), pointer :: ccdNodeNums(:)
    type(allFramesType), pointer :: ccdData(:)
    integer, allocatable :: previousEventFrame(:), numFrames(:)
    real(double), pointer :: eventTimes(:)
    character(2) :: ccdStr
    character(8) :: exposuTabName
    real(double) :: oldDeltaT, deltaT, timeAdvance
    real(double), allocatable :: avTimeDel(:)
logical :: testFlag = .false.

    !--------------------------------------------------------------------
    ! Start:

    call message(AppMsg, VERBOSE, 'Opening the event list.')

    eventSetName = stringParameter('eventset')
    set = dataSet(eventSetName, MODIFY)
    eventTab = table(set, 'EVENTS')
    numEvents = numberOfRows(eventtab)
    allocate(rands(0:numEvents-1))
    call getRandomNumbers(rands)
    rands = rands - 0.5
!**** BUT should check the TIMEPIXR kwd in both event list and each EXPOSU.

    ccdNodeNums => int8Data(column(eventTab, 'CCDNR'))
    minCcdNodeNum = minval(ccdNodeNums)
    maxCcdNodeNum = maxval(ccdNodeNums)
    allocate(ccdData(minCcdNodeNum:maxCcdNodeNum))
    allocate(previousEventFrame(minCcdNodeNum:maxCcdNodeNum))
    allocate(numFrames(minCcdNodeNum:maxCcdNodeNum))
    allocate(avTimeDel(minCcdNodeNum:maxCcdNodeNum))

    ccdData%ccdIsPresent = .false. ! default
    do trialCcdNodeNum = minCcdNodeNum, maxCcdNodeNum
      if (any(ccdNodeNums == trialCcdNodeNum))&
        ccdData(trialCcdNodeNum)%ccdIsPresent = .true.
    end do

    eventTimes => real64Data(column(eventTab, 'TIME'))

    !....................................................................
    call message(AppMsg, VERBOSE, 'Getting frame data.')

    do ccdNodeNum = minCcdNodeNum, maxCcdNodeNum
    if (.not. ccdData(ccdNodeNum)%ccdIsPresent) cycle
      write(ccdStr, '(i2.2)') ccdNodeNum
      exposuTabName = 'EXPOSU'//ccdStr
      call message(AppMsg, Verbose, "CCD/node number nn = "//ccdStr)

      exposuTab = table(set, exposuTabName)

      if (hasAttribute(exposuTab, 'RAND_TIM')) then
!**** should cater for mos as well!
        if (booleanAttribute(exposuTab, 'RAND_TIM')) then
        ccdData(ccdNodeNum)%ccdIsPresent = .false.
        call release(exposuTab)
    cycle
        end if
      end if

      call setAttribute(exposuTab, 'RAND_TIM', .true.)

      numFrames(ccdNodeNum) = numberOfRows(exposuTab)
      allocate(ccdData(ccdNodeNum)%frames(0:numFrames(ccdNodeNum)-1))
      call readColDataToFixed(exposuTab, 'TIME'&
        , ccdData(ccdNodeNum)%frames%time)

      if (hasAttribute(exposuTab, 'TIMEDEL')) then
        ccdData(ccdNodeNum)%frames%timeDel&
          = real64Attribute(exposuTab, 'TIMEDEL')
      else
        call readColDataToFixed(exposuTab, 'TIMEDEL'&
          , ccdData(ccdNodeNum)%frames%timeDel)
      end if

      call release(exposuTab)

      if (any(ccdData(ccdNodeNum)%frames(0:numFrames(ccdNodeNum)-2)%time&
        >= ccdData(ccdNodeNum)%frames(1:numFrames(ccdNodeNum)-1)%time))&
call error('disorderedFrames',"The frame times must occur in ascending &
&order for this task to work.") ! frames not in ascending order.

      avTimeDel(ccdNodeNum) = sum(ccdData(ccdNodeNum)%frames%timeDel)&
        / real(numFrames(ccdNodeNum), double)
    end do ! ccdNodeNum

    if (.not. any(ccdData%ccdIsPresent))&
call error('allRandomizedAlready',"No un-randomized times were found.")

    !....................................................................
    call message(AppMsg, VERBOSE, 'Dithering the event times.')

    previousEventFrame = 0
    do eventNum = 0, numEvents - 1
      errStr = sprintEvery(eventNum, numEvents, fraction=0.01)
      ! in ssclib/ssc_misc.
      if (errStr .ne. '') then
        call message(appMsg, verbose)
      end if

      ccdNodeNum = ccdNodeNums(eventNum)
    if (.not. ccdData(ccdNodeNum)%ccdIsPresent) cycle

if (testFlag) print *, 'eventNum', eventNum
if (testFlag) print *, 'ccdNodeNum', ccdNodeNum
if (testFlag) print *, 'previousEventFrame', previousEventFrame(ccdNodeNum)
      timeAdvance = eventTimes(eventNum)&
        - ccdData(ccdNodeNum)%frames(previousEventFrame(ccdNodeNum))%time

      approxNumFramesAdvanced = nint(timeAdvance / avTimeDel(ccdNodeNum))
!*** Better to get the frame duration from the cal (CAL_getModeParams) and divide by that.
      startFrameNum = previousEventFrame(ccdNodeNum) + approxNumFramesAdvanced
      startFrameNum = min(startFrameNum, numFrames(ccdNodeNum)-1)
      startFrameNum = max(startFrameNum, 0)
if (testFlag) print *, 'startFrameNum', startFrameNum
      oldDeltaT = abs(eventTimes(eventNum)&
        - ccdData(ccdNodeNum)%frames(startFrameNum)%time)

      if (eventTimes(eventNum)>ccdData(ccdNodeNum)%frames(startFrameNum)%time)&
      then
if (testFlag) print *, 'go higher'
        bestFrameNum = numFrames(ccdNodeNum) - 1 ! default
        do frameNum = startFrameNum + 1, numFrames(ccdNodeNum) - 1
if (testFlag) print *, '  frameNum', frameNum
          deltaT = abs(eventTimes(eventNum)&
            - ccdData(ccdNodeNum)%frames(frameNum)%time)

          if (deltaT > oldDeltaT) then
            bestFrameNum = frameNum
        exit
          end if

          oldDeltaT = deltaT
        end do

      else
if (testFlag) print *, 'go lower'
        bestFrameNum = 0 ! default
        do frameNum = startFrameNum - 1, 0, -1
if (testFlag) print *, '  frameNum', frameNum
          deltaT = abs(eventTimes(eventNum)&
            - ccdData(ccdNodeNum)%frames(frameNum)%time)

          if (deltaT > oldDeltaT) then
            bestFrameNum = frameNum
        exit
          end if

          oldDeltaT = deltaT
        end do
      end if

if (testFlag) print *, '  bestFrameNum', bestFrameNum
      if (oldDeltaT > ccdData(ccdNodeNum)%frames(bestFrameNum)%timeDel&
        / 2.0d0) then
if (testFlag) then
  print *, '     eventNum:', eventNum
  print *, '   ccdNodeNum:', ccdNodeNum
  print *, '    numFrames:', numFrames(ccdNodeNum)
  print *, 'startFrameNum:', startFrameNum
  print *, ' bestFrameNum:', bestFrameNum
  print *, '    oldDeltaT:', oldDeltaT
  print *, '    timeDel/2:', ccdData(ccdNodeNum)%frames(bestFrameNum)%timeDel/2.0d0
end if
call error('noFrameFound',"")
      end if

      eventTimes(eventNum) = ccdData(ccdNodeNum)%frames(bestFrameNum)%time&
        + ccdData(ccdNodeNum)%frames(bestFrameNum)%timeDel&
        * real(rands(eventNum), double)

      previousEventFrame(ccdNodeNum) = bestFrameNum
if (testFlag) print *, ''
    end do

    !....................................................................
    call message(AppMsg, VERBOSE, 'Releasing the event list dataset.')

    call release(set)

    do ccdNodeNum = minCcdNodeNum, maxCcdNodeNum
    if (.not. ccdData(ccdNodeNum)%ccdIsPresent) cycle
      deallocate(ccdData(ccdNodeNum)%frames)
    end do
    deallocate(ccdData)

  end subroutine etimedither

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module etimedither_mod
