!
! Module: evigmap_mod
!
! Authors:  Ian Stewart - SSC
!          
! Description: Makes an image in sky coordinates of just the vignetting function for the given energy range and instrument.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
!

module evigmap_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use cal
  use dal
  use dss		! for RangeT
  use param  
  use errorHandling
  use caloaldefs	! for SpacecraftAttitudeType

  use exposure_aux

  use binned_att

  use cal_aux
  use constants
  use coordinate
  use dal_aux
  use dss_ranges_aux	! for integrateRanges.
use dump
  use intervals_aux	! for OPEN_GTI etc.
  use save_image
  use wcs_aux

  implicit none 

  private 

  public evigmap

  ! Command-line parameters:
  character(256), allocatable ::&
    vigMapSetNames(:)
  character(256) ::&
    tempImageSetName = '',&
    templateImageSetName = '',&
    binnedAttSetName = ''
  type(RangeT), allocatable ::&
    userEeVRanges(:)
  integer ::&
    numUserRanges=0
  character(20) ::&
    attitudeStyle = ''

  integer(int32), parameter ::&
    numESamples = 50, numThetaSamples = 50, numPhiSamples = 36

contains 

  !---------------------------------------------------------------------
  subroutine evigmap
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: set
    type(ArrayT) :: imArray
    type(WcsT) :: wcs
    integer(int32) :: instrumentId
    real(double) ::&
      expStartTimeStamp, EeVsamples(numESamples)
!, startTimeFirstBin
    type(SpacecraftAttitudeType) :: medianScAttitude, dummyInstrumAttitude
    real(single), pointer :: templateImage(:,:)
    integer(int16) :: xiSize, yiSize
    type(RangeT) :: EeVRange
    type(attBinType), pointer :: binnedScAtt(:)
!    logical :: endsUnbounded(2)
    integer :: numAttBins, i, ei, ti, pi, attBinNum
!    type(IntervalT) :: dummyGti
    real(single), allocatable :: binnedAttWeights(:)
    real(double), allocatable ::&
      thetaArcsecSamples(:), phiSamples(:), tanX(:), tanY(:),&
      thetaMap(:,:), phiMap(:,:)
    real(single), allocatable ::&
      vigSamplesCube(:,:,:), vigSampleArrays(:,:,:), outCube(:,:,:),&
      xPixels(:), yPixels(:), tempVigMap(:,:)
    logical :: noWeightColumn

    real(double), parameter :: maxThetaArcsec = 1199.0d0

logical :: testFlag = .false.

    !--------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'aaa'

    call readParameters

    !...................................................................
    ! Read some essentials from the template image:

    set = dataSet(templateImageSetName, READ)
    imArray = array(set, 0)
if (testFlag) print *, 'bbb'
    call readWcs(set, imArray, wcs) ! in ssclib/wcs_aux.
if (testFlag) print *, 'ccc'

    call getCalInfo(set, instrumentId&
      , expStartTimeStamp, medianScAttitude, dummyInstrumAttitude)
    ! in ssclib/cal_aux.
if (testFlag) print *, 'ddd'

    call readArrayData(imArray, templateImage)
    ! in ssclib/dal_aux.
if (testFlag) print *, 'eee'

    call release(set)

    xiSize = size(templateImage, 1)
    yiSize = size(templateImage, 2)
    deallocate(templateImage)
if (testFlag) print *, 'fff'

!*** NOTE! the only things that the template image supplies are (i) wcs, (ii) image pixel dimensions, (iii) median attitude, (iv) cal instrument initialization  and (v) exposure start time stamp. It would be nice to have some facility to input the last of these in other ways.

    EeVRange%lower%value = minval(userEeVRanges%lower%value)
    EeVRange%upper%value = maxval(userEeVRanges%upper%value)
    EeVRange%lower%type = INCLUSIVE
    EeVRange%upper%type = INCLUSIVE
    call sampleEnergyRange(EeVRange, EeVsamples) ! in exposure_aux.
if (testFlag) print *, 'ggg'

    !...................................................................
    ! Construct the list of attitudes:

    select case (attitudeStyle)
    case('binnedset')
!      call readBinnedAttitude(binnedAttSetName, binnedScAtt&
!        , startTimeFirstBin, endsUnbounded)
!      ! in binned_att.
      call readBinnedAttitude(binnedAttSetName, 'ATT_BINS', binnedScAtt, noWeightColumn)
      ! in attcalc2/binned_att.
      numAttBins = size(binnedScAtt)

    case('template')
!      numAttBins = 1
!      allocate(binnedScAtt(1))
!      binnedScAtt(1)%att = medianScAttitude
!      binnedScAtt(1)%binEndTime = 0.0d0
!      binnedScAtt(1)%fracExp = 1.0
!      binnedScAtt(1)%isGood = .true.
!      startTimeFirstBin = 0.0d0
!      endsUnbounded = .true.
      numAttBins = 1
      allocate(binnedScAtt(1))
      call initializeBinnedAtt(binnedScAtt) ! in attcalc2/binned_att.
      binnedScAtt(1)%isGood = .true.
      binnedScAtt(1)%gtiDuration = 1.0d0 ! dummy value.
      binnedScAtt(1)%att = medianScAttitude

    case default
call error('badAttStyle',"The value of --attstyle was not recognized.")
    end select

if (testFlag) print *, 'hhh'

    ! I'm not going to bother weighting the attitude samples with gtis. Gtis tend to be ccd-specific; and the slowly-varying nature of the vignetting function makes it tolerant of many liberties. If the attitude jumps by a significant fraction of the field-of-view width then it is probably pointless to worry too much about vignetting anyway.
    !
    ! So anyway the following weighting just weights each sample by the sample duration fraction. Of course, this might lead to troubles if the ends are unbounded... hmm...

    allocate(binnedAttWeights(numAttBins))
    if (numAttBins > 1) then
!      dummyGti%lower = 0.0d0
!      dummyGti%upper = 0.0d0
if (testFlag) print *, 'iii'
!      call calcBinnedAttGtis(binnedScAtt, startTimeFirstBin&
!        , endsUnbounded, (/dummyGti/), OPEN_GTI, binnedAttWeights)
!      ! in binned_att.
binnedAttWeights = 1.0

if (testFlag) print *, 'jjj'

if (sum(binnedAttWeights) <= 0.0) call error('allWeightsZero',"")
      binnedAttWeights = binnedAttWeights / sum(binnedAttWeights)
    else
      binnedAttWeights = 1.0
    end if

    !....................................................................
    ! Set up the polar arrays of vignetting samples:

    allocate(thetaArcsecSamples(numThetaSamples))
    allocate(phiSamples(numPhiSamples))
    allocate(vigSamplesCube(numESamples, numThetaSamples, numPhiSamples))

    thetaArcsecSamples = (/(i/real(numThetaSamples),i=0,numThetaSamples-1)/)
    thetaArcsecSamples = thetaArcsecSamples * maxThetaArcsec
    phiSamples = (/(i/real(numPhiSamples-1),i=0,numPhiSamples-1)/)
    phiSamples = phiSamples * TwoPi

if (testFlag) print *, 'kkk'
    call constructVignettingCube(EeVsamples, thetaArcsecSamples&
      , phiSamples, vigSamplesCube) ! in exposure_aux.

if (testFlag) print *, 'lll'
    allocate(vigSampleArrays(numUserRanges, numThetaSamples, numPhiSamples))

    do ei = 1, numUserRanges
      do ti = 1, numThetaSamples
        do pi = 1, numPhiSamples
          vigSampleArrays(ei,ti,pi) = integrateRanges(real(EeVSamples, single)&
            , vigSamplesCube(:,ti,pi), userEeVRanges(ei))
          ! integrateRanges is in ssclib/dss_ranges_aux.
        end do
      end do
      vigSampleArrays(ei,:,:) = vigSampleArrays(ei,:,:)/vigSampleArrays(ei,1,1)
    end do
if (testFlag) print *, 'mmm'

    deallocate(vigSamplesCube)

    !...................................................................
    ! Allocate and initialize the cube of output images:

    allocate(outCube(numUserRanges, xiSize, yiSize))
    outCube = 0.0

    !...................................................................
    ! Now do the transforms:

    call message(AppMsg, verbose, "Transforming to sky coordinates.")

if (testFlag) print *, 'nnn'
    call CAL_setState(randomize=.false.)

    allocate(xPixels(xiSize), yPixels(yiSize))
    allocate(tanX(xiSize), tanY(yiSize))
    xPixels = (/(real(i,single), i=1,xiSize)/)
    yPixels = (/(real(i,single), i=1,yiSize)/)
if (testFlag) print *, 'ooo'
    call pixelsToTan(xPixels, yPixels, wcs, tanX, tanY) ! in ssclib/wcs_aux
if (testFlag) print *, 'ppp'

    allocate(tempVigMap(xiSize,yiSize))
    allocate(thetaMap(xiSize,yiSize), phiMap(xiSize,yiSize))
    do attBinNum = 1, numAttBins
    if (binnedAttWeights(attBinNum)==0.0) cycle

      call tanToTel(spread(tanX, 2, yiSize), spread(tanY, 1, xiSize)&
        , wcs%x%refWorld * PiOn180, wcs%y%refWorld * PiOn180&
        , binnedScAtt(attBinNum)%att&, expStartTimeStamp, thetaMap, phiMap)
      ! in ssclib/coordinate.

      where(thetaMap > maxThetaArcsec) thetaMap = maxThetaArcsec
!**** might want to flag such pixels???

if (testFlag) call dumpImageToFits(thetaMap, 'thetamap.ds')

      do ei = 1, numUserRanges
!print *, 'image ',ei
        write(errstr, '(a, i2.2)') 'Image ', ei
        call message(AppMsg, Verbose)

        call interpolateVigMapToArray(phiMap, thetaMap&
          , thetaArcsecSamples, phiSamples, vigSampleArrays(ei,:,:)&
          , tempVigMap) ! in exposure_aux.

        outCube(ei,:,:) = outCube(ei,:,:) + tempVigMap&
          * binnedAttWeights(attBinNum)
      end do
    end do
if (testFlag) print *, 'qqq'

    !...................................................................
    ! Save the output:

    call message(AppMsg, verbose, "Saving the output images.")

    do ei = 1, numUserRanges
      call saveImage(outCube(ei,:,:), templateImageSetName&
        , vigMapSetNames(ei), tempImageSetName)
      ! In ssclib/save_image.
    end do
    deallocate(binnedScAtt)

if (testFlag) print *, 'zzz'

  end subroutine evigmap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    integer :: rangeNum

    !-------------------------------------------------------------------
    ! Start:

    templateImageSetName = stringParameter('templateset')
    tempImageSetName = stringParameter('tempset')

    attitudeStyle = stringParameter('attstyle')
    select case(attitudeStyle)
    case('binnedset')
      binnedAttSetName = stringParameter('binnedattset')
    case('template')
    case default
call error('badAttStyle',"The value of --attstyle was not recognized.")
    end select

!    withBinnedAttSet = booleanParameter('withbinnedattset')
!    if (withBinnedAttSet) then
!      binnedAttSetName = stringParameter('binnedattset')
!    end if

    numUserRanges = parameterCount('evlo')
    if (numUserRanges <= 0)&
call error('EvLoListHasNoElements',"You must supply at least 1 &
&element in the list --evlo.")
    if (numUserRanges .ne. parameterCount('evhi'))&
call error('badNumberEvHi',"The parameter --evhi must have the &
&same number of elements as --evlo.")
    allocate(userEeVRanges(numUserRanges))
    userEeVRanges%lower%type = INCLUSIVE
    userEeVRanges%upper%type = INCLUSIVE
    do rangeNum = 1, numUserRanges
      userEeVRanges(rangeNum)%lower%value = realParameter('evlo', rangeNum-1)
      userEeVRanges(rangeNum)%upper%value = realParameter('evhi', rangeNum-1)
    end do

    if (numUserRanges .ne. parameterCount('outsets'))&
call error('badNumberOutsets',"The parameter --outsets must have the &
&same number of elements as --evlo.")

    allocate(vigMapSetNames(numUserRanges))
    do rangeNum = 1, numUserRanges
      vigMapSetNames(rangeNum) = stringParameter('outsets', rangeNum-1)
    end do

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module evigmap_mod


