!
! Module: lccorr_corrector
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains variables shared by more than 1 module.
! 
! $Id: lccorr_corrector.f90,v 1.2 2000/08/11 13:08:45 pvideler Exp $
!- 
module lccorr_corrector

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal	! for caltypes -> PsfBinSizeT
  use errorHandling
  use dss

  use lccorr_include
  use exposure_map
  use exposure_aux

  use array_utils	! for trapezoidSum
  use constants
  use coordinate	! for raDecToInst, instToRaw, getPsfImagePixelCorners
  use dal_aux		! for readColDataToFixed
  use dss_aux
  use dss_ranges_aux
!  use exposure
!use l_exposure
  use geometric_types	! for Point2dT and Point2dDbleT
  use intervals_aux	! for orIntervals
  use regridders	! for regridCartesian

use dump
!use plot_utils

  implicit none 

  private
 
  public calcCcdWeightsFromTheory
  public calcExposureCorrections
  public recordRateNullsAsGtis

contains 

  !---------------------------------------------------------------------
  subroutine calcCcdWeightsFromTheory(inSetName, timeAttWcs, expInfo&
  , requiredCorrections, ccdWeights, totalBkgWeight, srcRaDeg, srcDecDeg&
  , bkgMapSetName)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set to the correct instrument and filter before
  ! this routine is called.
  !---------------------------------------------------------------------
  !
  ! Global variables accessed:
  !	lowEnergyPsfEnergyFrac
  !	instrumentId
  !	pleaseCorrectDither
  !	numMicropixelDivs
  !	chipMinRawX
  !	chipMaxRawX
  !	chipMinRawY
  !	chipMaxRawY
  !
  !---------------------------------------------------------------------
  ! NOTE! That 'wcs' as fed to calcCcdExpMap below should be the wcs
  ! of the event list X and Y columns. This is needed to specify the
  ! X-Y coordinate system so that a mask from region filtering on these
  ! columns can be constructed.
  !---------------------------------------------------------------------

    implicit none

    character(*),       intent(in)           :: inSetName
    type(localTAWtype), intent(in)           :: timeAttWcs
    type(expInfoType),  intent(in)           :: expInfo(:,:)
    integer(int32),     intent(in)           :: requiredCorrections
    real(single),       intent(out)          :: ccdWeights(&
                                                  size(expInfo,1),&
                                                  size(expInfo,2)),&
                                                totalBkgWeight
    real(double),       intent(in), optional :: srcRaDeg, srcDecDeg
    character(*),       intent(in), optional :: bkgMapSetName

    integer(int8) :: numCcds, numNodes, ccdNum, nodeNum, ccdNodeNum
    logical(bool) :: isPointSource
!, allSpatialMapsBlank
    type(DataSetT) :: inSet, eventSet
    type(BlockT) :: rateBlock
    type(DssT) :: dssPointer
    type(RangeT) :: EeVrange
    real(double) :: EeVsamples(numEsamples),  srcThetaArcsec, srcPhi
    real(single) ::&
      spectrum(numEsamples), filterTrans(numEsamples), effArea(numEsamples),&
      EeVFunctions(numEsamples), vigEfunctionAtSrc(numEsamples),&
      qeArray(numCanonicalPatterns, numEsamples), totalQe(numEsamples),&
      eNorm, flatNorm, ccdFlatExp, ccdBkgWeight
    real(single), pointer ::&
      psfSamplesCubeTelcoord(:,:,:),&
      psfSamplesCubeCcd(:,:,:),&
      bkgSamplesCubeCcd(:,:,:)
    real(double), pointer :: psfPixelCnrsThetaArcsec(:,:), psfPixelCnrsPhi(:,:)
    type(PsfBinSizeT) :: psfPixelSizeMm
    real(single), allocatable :: ccdExpMap(:,:)
    character(2) :: ccdStr
    integer(int16) :: winCtrRawXi, winCtrRawYi, eI

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa calcCcdWeightsFromTheory'
    call message(AppMsg, Noisy, 'Entering subroutine &
      &calcCcdWeights.')

    numCcds  = size(expInfo,1)
    numNodes = size(expInfo,2)

    ! Construct outSet column names:

    isPointSource = .false. ! default
    if (present(srcRaDeg) .and. present(srcDecDeg)) then
      isPointSource = .true.
if (testFlag) print *, '  srcRaDeg', srcRaDeg
if (testFlag) print *, '  srcDecDeg', srcDecDeg
    end if

    inSet = dataSet(inSetName, READ)
    rateBlock = block(inSet, 'RATE')
    dssPointer = dataSubspace(rateBlock)

    !...................................................................
    ! Get the flux and response functions (including energy bounds and
    ! sample intervals):

if (testFlag) print *, '  bbb'

!!!    if (!** with spectrum) then
!!!      call getEnergyBounds(dssPointer, rateBlock, EeVrange, spectrum)
!!!    else
!      call getEnergyBounds_1(dssPointer, rateBlock, expInfo(1,1)%temperature&
!        , timeAttWcs%timeStamp, EeVrange) ! in exposure_map.

EeVrange = getEnergyBounds(dssPointer, rateBlock, expInfo(1,1)%temperature&
  , timeAttWcs%timeStamp)
! in exposure_map.

!!!    end if

if (testFlag) print *, 'EeVrange:', EeVrange%lower%value, 'to', EeVrange%upper%value
if (testFlag) print *, 'EeVrange bound types:', EeVrange%lower%type, EeVrange%upper%type

    call sampleEnergyRange(EeVrange, EeVsamples) ! in ssclib/exposure.
if (testFlag) print *, '  ccc'

!!!    if (!** with spectrum) then
!*** interpolate user spectrum
!!!    else
      call constructModelSpectrum(EeVsamples, spectrum) ! in exposure_map.
!!!    end if
if (testFlag) print *, '  ddd'

    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure.
if (testFlag) print *, '  eee'
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure.
if (testFlag) print *, '  fff'

    EeVFunctions = spectrum * effArea * filterTrans
if (testFlag) print *, '  ggg'

    if (isPointSource) then
      call raDecToTel(srcRaDeg*PiOn180, srcDecDeg*PiOn180, timeAttWcs%attitude&
       , timeAttWcs%timeStamp, srcThetaArcsec, srcPhi)
      ! in ssclib/coordinate.f90

      call calcVigAtSourcePos(srcThetaArcsec, srcPhi, EeVsamples&
        , vigEfunctionAtSrc) ! in ssclib/exposure.f90

      call CAL_SetState(accuracyLevel=ACCURACY_MEDIUM)

      call constructPsfCubeTelcoord(EeVsamples, srcThetaArcsec, srcPhi&
        , lowEnergyPsfEnergyFrac, psfSamplesCubeTelcoord, psfPixelSizeMm)
      ! in ssclib/exposure.

if (testFlag) print *, '  hhh'
if (testFlag) call dumpImageToFits(psfSamplesCubeTelcoord(1,:,:), 'psf.ds')

      allocate(psfPixelCnrsThetaArcsec(&
        size(psfSamplesCubeTelcoord,2)+1,&
        size(psfSamplesCubeTelcoord,3)+1)&
      )
      allocate(psfPixelCnrsPhi(&
        size(psfSamplesCubeTelcoord,2)+1,&
        size(psfSamplesCubeTelcoord,3)+1)&
      )
      call getPsfPixelCnrsThetaPhi(srcThetaArcsec, srcPhi&
        , psfPixelSizeMm, psfSamplesCubeTelcoord(1,:,:)&
        , psfPixelCnrsThetaArcsec, psfPixelCnrsPhi)
      ! in ssclib/coordinate.

!***** write psf array size to output table

!**** check Hermann's claim that the telcoord coords are inverted

      EeVFunctions = EeVFunctions * vigEfunctionAtSrc

    else
      vigEfunctionAtSrc = 1.0
    end if
if (testFlag) print *, '  iii'

    !...................................................................
    ! Other init:

    allocate(ccdExpMap(chipMinRawX:chipMaxRawX, chipMinRawY:chipMaxRawY))

call message(AppMsg, Verbose, 'Opening the event set.')
eventSet = dataSet(eventSetName, READ)

!    allSpatialMapsBlank = .true. ! default
    totalBkgWeight = 0.0
    ccdWeights = 0.0

if (testFlag) print *, '  jjj'
    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
      if (.not. expInfo(ccdNum, nodeNum)%exposuExists) cycle
!if (testFlag .and. ccdNum .ne. 3) cycle

        if (instrumentId == EPN) then ! just to stop the cal giving a warning:
          call CAL_setState(ccdChipId=int(ccdNum, int32))
        else
          call CAL_setState(ccdChipId=int(ccdNum, int32)&
            , ccdNodeId=int(nodeNum - 1, int32))
        end if
if (testFlag) print *, '  kkk'

        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum
        write(ccdStr, '(i2.2)') ccdNodeNum
        call message(AppMsg, Verbose, "********************************")
        call message(AppMsg, Verbose, "Found EXPOSUnn extension for &
          &CCD/node number nn = "//ccdStr)
if (testFlag) print *, '  lll'

        winCtrRawXi = nint((expInfo(ccdNum,nodeNum)%winMinRawX&
          + expInfo(ccdNum,nodeNum)%winMaxRawX) / 2.0)
        winCtrRawYi = nint((expInfo(ccdNum,nodeNum)%winMinRawY&
          + expInfo(ccdNum,nodeNum)%winMaxRawY) / 2.0)
if (testFlag) print *, '  mmm'

        call constructQeArray(EeVsamples, qeArray, winCtrRawXi, winCtrRawYi)
        ! in ssclib/exposure.f90
if (testFlag) print *, '  nnn'

        eNorm = calcEeVFunctionNorm(EeVsamples, spectrum * effArea, filterTrans&
          , vigEfunctionAtSrc, qeArray, requiredCorrections) ! in exposure_map
if (testFlag) print *, '  ooo'

        totalQe = (/(sum(qeArray(:,eI)), eI=1,numESamples)/)
        flatNorm = integrateRanges(real(EeVSamples, single)&
          , EeVfunctions * totalQe, (/EeVrange/))

        !...............................................................
        ! The first step is to call calcCcdExpMap. This subroutine takes a
        ! pointer to the data subspace, a spectrum and (optionally) a flux
        ! cube, and returns both an exposure map and a 'flat exposure'
        ! value. The latter is the result of an internal integration in
        ! the spatial coordinates of a second exposure map formed under
        ! the assumption that the event flux is spatially uniform. This
        ! can be useful for estimating background exposure if the background
        ! map is not known.
        !
        ! The aim of this part is to calculate two numbers: the actual
        ! exposure weight for this CCD, and the weight that would have
        ! resulted in the case of a background event flux. (If the time
        ! series is the background one then these two numbers will have
        ! the same value; if the TS is from the source itself, then they
        ! will be different.) The exposure weight is needed for exposure
        ! correction; the background exposure weight will be needed
        ! externally to scale the background TS to the same collection
        ! area as the source TS.

        if (isPointSource) then
if (testFlag) print *, '  ppp'
          call interpolatePsfCubeToCcd(psfPixelCnrsPhi, psfPixelCnrsThetaArcsec&
            , psfSamplesCubeTelcoord, psfSamplesCubeCcd)!, testFlagArg=ptf)

          ! Note to later developers: DON'T be tempted to dump out of the
          ! loop at this point if no part of the psf overlaps the chip.
          ! We still need to see if any of the spatial selection region
          ! overlaps the chip in order to calculate the BKGRATIO value.
if (testFlag) print *, '  qqq'

          call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
            , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
            , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
            , psfSamplesCubeCcd, ccdExpMap, ccdFlatExp&
          ) ! in exposure_map.

if (testFlag) call error('testHalt',"test flag set; stopping here.")

          deallocate(psfSamplesCubeCcd)
          ccdWeights(ccdNum,nodeNum) = sum(ccdExpMap) / eNorm

          if (present(bkgMapSetName)) then
!*** call getBkgCube(bkgMapSetName, ccdNodeNum, bkgSamplesCubeCcd)
!*** also supply instrumentId
call error('',"")

            call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
              , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
              , bkgSamplesCubeCcd, ccdExpMap, ccdFlatExp&
            ) ! in exposure_map.

            deallocate(bkgSamplesCubeCcd)
            ccdBkgWeight = sum(ccdExpMap) / eNorm !*** / flatNorm instead??
          else
            ccdBkgWeight = ccdFlatExp / flatNorm
          end if

        else ! this is the bkg TS not the source TS.
          if (present(bkgMapSetName)) then
!*** call getBkgCube(bkgMapSetName, ccdNodeNum, bkgSamplesCubeCcd)
!*** also supply instrumentId
call error('',"")

            call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
              , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
              , bkgSamplesCubeCcd, ccdExpMap, ccdFlatExp&
            ) ! in exposure_map.

            deallocate(bkgSamplesCubeCcd)
            ccdWeights(ccdNum,nodeNum) = sum(ccdExpMap) / eNorm
            ccdBkgWeight = ccdWeights(ccdNum,nodeNum)!*** / flatNorm instead??

          else
            call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
              , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
              , ccdExpMap=ccdExpMap, ccdFlatExp=ccdFlatExp&
            ) ! in exposure_map.

            ccdWeights(ccdNum,nodeNum) = ccdFlatExp / eNorm
            ccdBkgWeight = ccdFlatExp / flatNorm
          end if
        end if
        totalBkgWeight = totalBkgWeight + ccdBkgWeight
if (testFlag) print *, 'ccdFlatExp, flatNorm:', ccdFlatExp, flatNorm
if (testFlag) print *, 'ccdBkgWeight, total bkgWeight:', ccdBkgWeight, totalBkgWeight

      end do ! loop over nodes
    end do ! loop over ccds
if (testFlag) print *, '  rrr'

call release(eventSet)

    call release(inSet)
    call dssRelease(dssPointer)

    call message(AppMsg, Verbose, "********************************")

    if (isPointSource) then
      deallocate(psfSamplesCubeTelcoord)
      deallocate(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi)
    end if
if (testFlag) print *, '  zzz calcCcdWeightsFromTheory'

  end subroutine calcCcdWeightsFromTheory

  !---------------------------------------------------------------------
  subroutine calcExposureCorrections(inSetName, outSetName, eventSetName&
  , expInfo, requiredCorrections, isPointSource, applyCorrections, ccdWeights)
  !---------------------------------------------------------------------

    implicit none

    character(*),      intent(in) :: inSetName, outSetName, eventSetName
    type(expInfoType), intent(in) :: expInfo(:,:)
    integer(int32),    intent(in) :: requiredCorrections
    logical,           intent(in) :: isPointSource
    logical(bool),     intent(in) :: applyCorrections
    real(single),      intent(in) :: ccdWeights(&
                                       size(expInfo,1),&
                                       size(expInfo,2))

    integer(int8) :: numCcds, numNodes, ccdNum, nodeNum, ccdNodeNum
    character(8) ::&
      rateColName, errorColName, fracExpColName, outGtiTabName, exposuTabName
    type(DataSetT) :: inSet, outSet, eventSet
    type(BlockT) :: rateBlock
    type(DssT) :: dssPointer
    integer(int32) :: numRows, row
    real(single), allocatable ::&
      fracExp(:), ccdFracExp(:), ccdAntiAlias(:), antiAlias(:)
    logical(bool), allocatable :: badBins(:)
    logical(bool) :: someEventsNotRandomized
    type(IntervalT), pointer :: gtis(:)
    character(2) :: ccdStr
    type(TableT) :: outTab
    real(single), pointer :: tsFracExp(:), rate(:), uncert(:)
    type(ColumnT) :: rateCol, errorCol
    real(single) :: totalCcdWeight
integer :: i
logical(bool) ::&
  treatGti, treatCosmicRays, treatDeadTime, treatAlias
!**** Neaten this up some time.

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Noisy, 'Entering subroutine &
      &calcExposureCorrections.')

    numCcds  = size(expInfo,1)
    numNodes = size(expInfo,2)

treatGti         = btest(requiredCorrections, TREAT_GTI_BIT)
treatCosmicRays  = btest(requiredCorrections, TREAT_COSMIC_BIT)
treatDeadTime    = btest(requiredCorrections, TREAT_DEAD_TIME_BIT)
treatAlias       = btest(requiredCorrections, TREAT_ALIAS_BIT)
!**** Neaten this up some time. ALSO! Shouldn't use names of globals.

    ! Construct outSet column names:

    if (isPointSource) then
      rateColName    = 'RATE'
      errorColName   = 'ERROR'
      fracExpColName = 'FRACEXP'
      outGtiTabName  = 'SRC_GTIS'
    else
      rateColName    = 'BRATE'
      errorColName   = 'BERROR'
      fracExpColName = 'BFRACEXP'
      outGtiTabName  = 'BKG_GTIS'
    end if
if (testFlag) print *, 'col/ext names: '&
//trim(rateColName)//' '//trim(errorColName)//' '//trim(fracExpColName)&
//' '//trim(outGtiTabName)

    inSet = dataSet(inSetName, READ)
    rateBlock = block(inSet, 'RATE')
    dssPointer = dataSubspace(rateBlock)

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')
    numRows = numberOfRows(outTab)
    allocate(fracExp(numRows))
    call readColDataToFixed(outTab, fracExpColName, fracExp)
    call release(outSet)

    fracExp = 0.0
    allocate(ccdFracExp(numRows), ccdAntiAlias(numRows), antiAlias(numRows))
    allocate(badBins(numRows))
    antiAlias = 0.0
    badBins = .false.

    someEventsNotRandomized = .false. ! default
!    allocate(oredGtis(0))

if (testFlag) print *, 'hhh'
    eventSet = dataSet(eventSetName, READ)
    totalCcdWeight = 0.0 ! default
    do ccdNum = 1, numCcds ! numCcds, numNodes are declared in lccorr_include
      do nodeNum = 1, numNodes ! and their values set in miscSetups.
      if (ccdWeights(ccdNum, nodeNum) <= 0.0) cycle

        if (.not. expInfo(ccdNum, nodeNum)%exposuExists) then
          write(errStr, '(a, i2.2, a, i1, a)') "The weight for CCD "&
            , ccdNum, " node ", nodeNum-1, " is non-zero, but there is &
            &no corresponding EXPOSU extension."
          call warning('badWeightList')
      cycle
        end if

        totalCcdWeight = totalCcdWeight + ccdWeights(ccdNum, nodeNum)

        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum
        write(ccdStr, '(i2.2)') ccdNodeNum
        exposuTabName = 'EXPOSU'//ccdStr

        call message(AppMsg, Verbose, "Calculating time-dependent factors &
          &for CCD/node number nn = "//ccdStr)

        write(errstr, '(a, f6.4)') "Weight for this CCD is "&
          , ccdWeights(ccdNum, nodeNum)
        call message(AppMsg, Verbose)

        ccdFracExp   = 1.0 ! default
        ccdAntiAlias = 1.0 ! default

        if (hasTimeFilter(dssPointer, ccdNodeNum)) then
!**** check that the time filter is of gti type.
!          call readGtis(inSet, rateBlock, int(ccdNodeNum), gtis)
call readGtisFromDss(inSet, rateBlock, int(ccdNodeNum), gtis)

          call correctForGtiAndAlias(eventSet, exposuTabName, outSetName&
            , gtis, treatGti, treatCosmicRays, treatDeadTime, treatAlias&
            , expInfo(ccdNum, nodeNum)%tFrame, ccdFracExp, ccdAntiAlias&
            , badBins)
          ! in exposure_aux.

          deallocate(gtis)
        end if ! hasTimeFilter

if (testFlag) then
  print *, 'ccdNodeNum', ccdNodeNum
  print *, 'fracExp, ccdFracExp, ccdWeights:', fracExp(1), ccdFracExp(1), ccdWeights(ccdNum, nodeNum)
  print *, 'antiAlias, ccdAntiAlias, ccdWeights:', antiAlias(1), ccdAntiAlias(1), ccdWeights(ccdNum, nodeNum)
  if (any(ccdFracExp > 1.0)) then
    i = smallestTrueIndex(ccdFracExp > 1.0)
    print *, '!!!!! ccdFracExp(',i,')=',ccdFracExp(i)
  end if
end if

        fracExp   = fracExp   + ccdFracExp   * ccdWeights(ccdNum, nodeNum)
        antiAlias = antiAlias + ccdAntiAlias * ccdWeights(ccdNum, nodeNum)

        if (any(ccdAntiAlias < 1.0)) someEventsNotRandomized = .true.
!*** check that this works.
      end do ! loop over nodes
    end do ! loop over ccds
    call release(eventSet)

!    if (.not. isPointSource) then
!      fracExp = fracExp / bkgWeight
!      antiAlias = antiAlias / bkgWeight
!    end if

    call release(inSet)
    call dssRelease(dssPointer)

    if (totalCcdWeight <= 0.0)&
call error('bug5',"Please contact the code developer.")

    fracExp   = fracExp   / totalCcdWeight
    antiAlias = antiAlias / totalCcdWeight

    where(badBins) fracExp = 0.0
if (testFlag) print *, 'count(badBins)=', count(badBins)

    if (someEventsNotRandomized) then
      fracExp = fracExp * antiAlias
    end if

    !....................................................................
    ! Apply exposure corrections:

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')
    tsFracExp => real32Data(column(outTab, fracExpColName))
    tsFracExp = fracExp
if (testFlag) print *, 'count(tsFracExp<=0)=', count(tsFracExp<=0)

    if (applyCorrections) then
      call message(AppMsg, Verbose, 'Applying exposure corrections.')

      rateCol  = column(outTab, rateColName)
      errorCol = column(outTab, errorColName)
      rate  => real32data(rateCol)
      uncert => real32data(errorCol)

      do row = 0, numRows - 1
        if (tsFracExp(row) > 0.0) then
          rate(row)   = rate(row)   / tsFracExp(row)
          uncert(row) = uncert(row) / tsFracExp(row)
        else
if (testFlag) print *, 'set to null row', row
          call setToNull(rateCol,  row)
          call setToNull(errorCol, row)
        end if
      end do ! next row.
    end if ! applyCorrections.

!    if (size(oredGtis) > 0) then
!      outGtiTab = addTable(outSet, outGtiTabName, size(oredGtis))
!      startVals => real64Data(addColumn(outGtiTab, 'START', REAL64, 's'))
!      stopVals  => real64Data(addColumn(outGtiTab, 'STOP',  REAL64, 's'))
!      startVals = oredGtis%lower
!      stopVals  = oredGtis%upper
!    end if

    call release(outSet)

  end subroutine calcExposureCorrections

  !---------------------------------------------------------------------
  subroutine interpolatePsfCubeToCcd(psfPixelCnrsPhi, psfPixelCnrsThetaArcsec&
  , psfSamplesCubeTelcoord, psfSamplesCubeCcd, doDummy)
!, testFlagArg)
  !---------------------------------------------------------------------
  ! NOTE that this routine requires that the cal have been set up by the
  ! calling routine for the correct ccd, node number and instrument.
  !
  ! NOTE psfSamplesCubeCcd should be deallocated after use.

!*** can a coordinate routine be used??

    implicit none 

    real(double),      intent(in)  :: psfPixelCnrsPhi(:,:),&
                                      psfPixelCnrsThetaArcsec(&
                                        size(psfPixelCnrsPhi,1),&
                                        size(psfPixelCnrsPhi,2)&
                                      )
    real(single),      intent(in)  :: psfSamplesCubeTelcoord(:,:,:)

    real(single),      pointer     :: psfSamplesCubeCcd(:,:,:)
    logical, intent(in), optional  :: doDummy
!, testFlagArg

!    type(Point2dT) ::&
!      psfImgPixCrnrRaw(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2))
type(Point2dDbleT), allocatable :: psfImgPixCrnrRaw(:,:)
type(Point2dT), allocatable :: psfImgPixCrnrSng(:,:)
    ! These data types are defined in ssclib/geometric_types.

    integer(int16) :: minRawXi, maxRawXi, minRawYi, maxRawYi, numEsamples, eI
    integer :: foreignStatus
    logical(bool), allocatable :: isOffChip(:,:)
    real(single) :: norm
!real(double), allocatable :: dummyX(:,:), dummyY(:,:)

logical :: testFlag = .false.
!integer :: loxi, hixi, loyi, hiyi

    !-------------------------------------------------------------------
    ! Start:

!if (present(testFlagArg)) then
!  testFlag = testFlagArg
!else
!  testFlag = .false.
!end if

if (testFlag) print *, '  aaa interpolatePsfCubeToCcd'

    call message(AppMsg, Verbose, "  Converting PSF cube to chip &
      &coordinates for this CCD:")

    if ( size(psfPixelCnrsPhi,1) .ne. size(psfSamplesCubeTelcoord,2)+1&
    .or. size(psfPixelCnrsPhi,2) .ne. size(psfSamplesCubeTelcoord,3)+1)&
call error('bug2',"Please contact the code developer.")

allocate(psfImgPixCrnrRaw(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)))
allocate(psfImgPixCrnrSng(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)))
allocate(isOffChip(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)))

    numEsamples = size(psfSamplesCubeTelcoord, 1)

    ! Find the positions of the PSF-array pixel corners in the chip frame:

    call message(AppMsg, Verbose, "    Converting PSF grid theta-phi &
      &values to RAWXY.")

if (testFlag) print *, '  bbb0'
    call telToRaw(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi&
     , psfImgPixCrnrRaw%x, psfImgPixCrnrRaw%y, isOffChip)
if (testFlag) then
  print *, size(psfPixelCnrsThetaArcsec,1), size(psfPixelCnrsThetaArcsec,2)
  print *, size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)
  print *, size(psfImgPixCrnrRaw%x,1), size(psfImgPixCrnrRaw%x,2)
  print *, size(psfImgPixCrnrRaw%y,1), size(psfImgPixCrnrRaw%y,2)
  print *, size(isOffChip,1), size(isOffChip,2)
end if
!!!call telToRawArray(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi&
!!!     , psfImgPixCrnrRaw%x, psfImgPixCrnrRaw%y, isOffChip)
!allocate(dummyX(size(psfPixelCnrsPhi,1),size(psfPixelCnrsPhi,2)))
!allocate(dummyY(size(psfPixelCnrsPhi,1),size(psfPixelCnrsPhi,2)))
!dummyX = psfImgPixCrnrRaw%x
!dummyY = psfImgPixCrnrRaw%y
!call telToRawArray(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi&
!     , dummyX, dummyY, isOffChip)
    ! in ssclib/coordinate.
if (testFlag) print *, '  bbb1'

    ! Now find out how big the chip-coord psf image array will have to be:

    minRawXi = nint(minval(psfImgPixCrnrRaw%x))
    maxRawXi = nint(maxval(psfImgPixCrnrRaw%x))
    minRawYi = nint(minval(psfImgPixCrnrRaw%y))
    maxRawYi = nint(maxval(psfImgPixCrnrRaw%y))

if (testFlag) print *, '  ccc'

    if (minRawXi > maxRawXi .or. minRawYi > maxRawYi) then
call error('bug3',"Please contact the code developer.")
!      allocate(psfSamplesCubeCcd(numEsamples, 0, 0))
!  return
    end if
if (testFlag) print *, '  ddd'

    allocate(psfSamplesCubeCcd(numEsamples&
      , minRawXi:maxRawXi, minRawYi:maxRawYi))
if (testFlag) print *, '  eee'

    if (present(doDummy)) then
if (testFlag) print *, '  fff'
      if (doDummy) then
if (testFlag) print *, '  ggg'
        call warning('dummyPsfCube', "This is a dummy routine! It doesn't &
          &return the real psfSamplesCubeCcd.")
if (testFlag) print *, '  hhh denom=', real((1+maxRawXi-minRawXi)*(1+maxRawYi-minRawYi))
        psfSamplesCubeCcd = 1.0&
          / real((1+maxRawXi-minRawXi)*(1+maxRawYi-minRawYi))
if (testFlag) print *, '  iii'
  return
      end if
    end if

    psfSamplesCubeCcd = 0.0

    if (   minRawXi > chipMaxRawX .or. maxRawXi < chipMinRawX&
      .or. minRawYi > chipMaxRawY .or. maxRawYi < chipMinRawY) then

      call message(AppMsg, Verbose, "    No significant overlap of PSF &
        &onto the chip. Calculate selection area only.")
  return
    end if

    ! Final step before the regridding is to normalize the corner coordinates
    ! to the pixel frame of psfSamplesCubeCcd:

!    psfImgPixCrnrRaw%x = psfImgPixCrnrRaw%x - real(minRawXi - 1)
!    psfImgPixCrnrRaw%y = psfImgPixCrnrRaw%y - real(minRawYi - 1)
psfImgPixCrnrSng%x = psfImgPixCrnrRaw%x - real(minRawXi - 1, single)
psfImgPixCrnrSng%y = psfImgPixCrnrRaw%y - real(minRawYi - 1, single)

    call message(AppMsg, Verbose, "    Rebinning PSF image pixel values to the &
      &chip grid.")

    ! Now let us regrid:
!    do eI = 1, numEsamples
!if (testFlag) then
!  print *, ''
!  print *, '    jjja eI=', eI
!end if

if (testFlag) print *, '    jjja'

call regridCartesian(psfImgPixCrnrSng, psfSamplesCubeTelcoord&
  , psfSamplesCubeCcd, foreignStatus, logical(testFlag, bool))
      ! in ssclib/regridders.
if (testFlag) print *, '    jjjb'

      if (foreignStatus < 0) then
print *, foreignStatus
call error('badRegridStatus',"The subroutine regrid returned with a bad &
  &status. You should contact the code developer.")
      end if
if (testFlag) print *, '    jjjc'

do eI = 1, numEsamples
      norm = sum(psfSamplesCubeCcd(eI,:,:))
if (testFlag) print *, '    jjjd'
if (norm <= 0) call error('bug4',"Please contact the code developer.")
if (testFlag) print *, '    jjje'
      psfSamplesCubeCcd(eI,:,:) = psfSamplesCubeCcd(eI,:,:) / norm
if (testFlag) then
  print *, '    jjjf'
  if (eI == 1) call dumpImageToFits(psfSamplesCubeCcd(eI,:,:), 'psf_chip.ds')
end if
    end do

if (testFlag) print *, '  zzz interpolatePsfCubeToCcd'

  end subroutine interpolatePsfCubeToCcd

  !---------------------------------------------------------------------
  subroutine recordRateNullsAsGtis(outSetName, rateColName, outGtiTabName)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: outSetName, rateColName, outGtiTabName

    type(DataSetT) :: outSet
    type(TableT) :: rateTab, outGtiTab
    type(ColumnT) :: col
    integer :: numRows, row, numGtis, gtiNum, startRow, finisRow
    logical, allocatable :: rateIsNull(:)
    integer, allocatable :: startRows(:), finisRows(:)
    logical :: withinGti
    real(double), pointer :: binCentreTime(:), startVals(:), stopVals(:)
    real(single), pointer :: timeDel(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    outSet = dataSet(outSetName, MODIFY)
    rateTab = table(outSet, 'RATE')
    numRows = numberOfRows(rateTab)

    binCentreTime => real64Data(column(rateTab, 'TIME'))

    allocate(timeDel(0:numRows-1))
    if (hasColumn(rateTab, 'TIMEDEL')) then
      call readColDataToFixed(rateTab, 'TIMEDEL', timeDel)
    else
      timeDel = real64Attribute(rateTab, 'TIMEDEL')
    end if

    col = column(rateTab, rateColName)
    allocate(rateIsNull(numRows))
    rateIsNull = .false. ! default
    do row = 0, numRows - 1
      if (isNull(col, row)) rateIsNull(row+1) = .true.
    end do

    ! Calculate the start and end rows of stretches of TRUE values:

    allocate(startRows(numRows), finisRows(numRows))
    startRows = 0
    finisRows = 0

if (testFlag) print *, ''
if (testFlag) print *, ''
    numGtis = 0
    withinGti = .false.
    do row = 1, numRows
      if (rateIsNull(row)) then
        if (withinGti) then
          finisRows(numGtis) = row - 1
          withinGti = .false.
if (testFlag) print *, '    rateIsNull,     withinGti. row=', row
        else
if (testFlag) print *, '    rateIsNull, NOT withinGti. row=', row
        end if
      else
        if (.not.withinGti) then
          numGtis = numGtis + 1
          startRows(numGtis) = row
          withinGti = .true.
if (testFlag) print *, 'NOT rateIsNull, NOT withinGti. row=', row
        else
if (testFlag) print *, 'NOT rateIsNull,     withinGti. row=', row
        end if
      end if
    end do
    if (withinGti) then
      finisRows(numGtis) = numRows
    end if
if (testFlag) print *, ''
if (testFlag) print *, ''

    outGtiTab = addTable(outSet, outGtiTabName, numGtis)
    startVals => real64Data(addColumn(outGtiTab, 'START', REAL64, 's'))
    stopVals  => real64Data(addColumn(outGtiTab, 'STOP',  REAL64, 's'))
    do gtiNum = 0, numGtis - 1
      startRow = startRows(gtiNum+1)-1
      finisRow = finisRows(gtiNum+1)-1
if (testFlag) print *, 'gtiNum, startRow, finisRow:', gtiNum+1, startRow+1, finisRow+1
      startVals(gtiNum) = binCentreTime(startRow) - timeDel(startRow) / 2.0d0
      stopVals(gtiNum)  = binCentreTime(finisRow) + timeDel(finisRow) / 2.0d0
    end do
if (testFlag) print *, ''
if (testFlag) print *, ''

    call setAttribute(block(outGtiTab), attribute(rateTab, 'MJDREF'))
    if (hasAttribute(rateTab, 'TIMEZERO')) then
      call setAttribute(block(outGtiTab), attribute(rateTab, 'TIMEZERO'))
    else
      call setAttribute(outGtiTab, 'TIMEZERO', 0.0d0, 's'&
        , "It's necessary to include this until SOC SPR-2565 is fixed.")
    end if

    call release(outSet)
    deallocate(timeDel)

  end subroutine recordRateNullsAsGtis

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module lccorr_corrector
